{ Except.pas   v3.0  Copyright  5/94  Adolfo Di Mare }

{ The following code is copyrighted by Adolfo Di Mare.
  - You can use it at your own risk,
  - but you cannot sell it.

                              Copyright
                              Adolfo Di Mare
                              Univ. Costa Rica
                              adolfo@di-mare.com
                              FAX: (506) 438-0139
                              Phn: (506) 253-4853
}

{$I ADT-ADH.inc}

UNIT Except;
{
> PROGRAMMER:      ADH: Adolfo Di Mare

> COLABORATORS:    FEC: Fabio Esquivel Chacn
                   WM:  William Martnez
                   GAS: Gustavo Alonso Sanabria

> RESULT
  Implements exception handling, following Ada's
  Exception/Raise paradigm, extended with the Signal()
  primitive.  The return point is set by procedure
  Exception() and the other procedures jump to the
  corresponding exception handler when invoked.

  The status code that corresponds to the exception is
  returned by function Catch() which clears it after
  being called.

> USAGE
  The general scheleton of a program block which
  includes an exception handler is the following:    }

{$IFDEF COMMENT}

    IF NOT Exception THEN BEGIN
      a := Exp(100000);        { Blows up            }
      Cicle(10);               { Never executed      }
      Leave_Exception_Context; { Pop exception stack }
    END
    ELSE BEGIN                 { Exception handler   }
      CASE Catch OF
        Program_Stack_Overflow :
          a := 0;
        User_Defined_Exception:
          WriteLn('Ignored by Signal');
        Floating_Point_Overflow,
        Floating_Point_Underflow :
          a := 1.e+37;
      ELSE
        Re_Throw;              { Branch to enclosing }
      END                      { exception handler   }
    END;

{$ENDIF}

{ NOTES
  An exception stack is kept to return to the correcto
  exception handler. The programmer doesn't need to
  keep track of where to return, as is the case when
  using Set_Jump() and Long_Jump() alone. In [Lee-83]
  a good C exception package is presented.  It can't
  be emulated here because Pascal lacks a
  preprocessor.

  The price of using exceptions is paid when an 
  exception occurs and not when an exception context is 
  entered.  This means that the cost of executing the 
  Exception() procedure is always O(1), but when an 
  exception occurs O(n) work could be incurred,
  where n is the size of the exception stack.

> REQUIRES
  This unit forces itself to be the first in Turbo's 
  ExitProc procedure chain. If any other unit needs to
  be the first, then the result will almost for sure
  be that the program will hang. This is required
  because the purpose of Except is to catch most run 
  time error conditions.  However, a savvy programmer 
  could modify Except to let other units take control 
  before it.

  Register_Exit_Procedure() supports the use of exit 
  procedures.

  The programmer needs to keep poping the stack upon 
  exit of every procedure that contains an exception
  handler, using the Leave_Exception_Context() and/or 
  Release_Nested_Contexts() procedures. Regretfully, 
  there is NO WAY to avoid forcing the programmer to
  pop the exception stack, and when this happens, the
  program bombs in a VERY UGLY way.

> ACKNOWLEDGEMENTS

  David Chaves:
  David did A LOT of the initial assembly coding.
  Later, even Adolfo had to step in.

  William Martnez:
  William started to debug the package, just after
  Adolfo had to quit doing it.  William also figured
  out a way to lift the requirement of putting Except
  as the last USES clause in the main program.

  Gustavo Alonso Sanabria:
  Alonso took Adolfo+William's work and ideas and
  debugged it. He did so in the semester break in 
  1991.  Gustavo deviced the Leave_Exception_Context 
  and Release_Nested_Contexts procedures, as the
  original scheme used by Adolfo was very cumbersome
  to use:
  - now the programmer can USE the Except unit in any
    place withing the USES clause (William Martinez
    figured out the trick for this). In the past,
    Except had to be the very last unit used in a
    program.
  - Adolfo got rid of Release_Nested_Contexts, because
    its use promotes bad programming habits.

  The Release_Nested_Contexts() procedure was
  introduced to avoid clearing many nested exception
  blocks, which requires invocking procedure
  Leave_Exception_Context() many times.

  Fabio Esquivel Chacn:
  Fabio debugged the Exception() procedure to force it
  to work under Turbo v6.0. Previous versions had a
  sub ax, =6, and Fabio discovered, after many long
  hours of work, that an add was needed. What is
  astonishing is that this bug was not found for at
  least one year (because of Turbo v5.5's way of
  linkin to a subroutine).


  [Lee-83] P.A. Lee: Exception Handling in C Programs;
     Software Practice and Experience; Vol 13; p389-
     405; 1983.

  Mr. Lee wrote the paper that inspired Adolfo to
  develope the exception package. This paper should be
  required reading for any [C && C++] programmer!
  Pascal programmer should also read it.
}

{[]=========[]}
{} INTERFACE {}
{[]=========[]}

TYPE
  PExitProc   = PROCEDURE;
  PDestructor = PROCEDURE(VAR ADT_instance);

  PException = ^PException;
  TException = RECORD
    Rep : ARRAY[1..4*SizeOf(POINTER)] OF BYTE;
  END;

FUNCTION  Exception_Code    : INTEGER;
FUNCTION  Catch             : INTEGER;
FUNCTION  Exception         : BOOLEAN;
PROCEDURE Throw       (code : INTEGER);
PROCEDURE Signal      (code : INTEGER);
PROCEDURE Re_Throw;
PROCEDURE Leave_Exception_Context;

PROCEDURE Register_Exit_Procedure(
            VAR ExitSave: PExitProc;
                ExitProc: PExitProc);
FUNCTION  Exception_Depth   : INTEGER;

PROCEDURE Register_ADT(  VAR obj;
                         VAR snode: TException;
                             done : POINTER);
PROCEDURE UnRegister_ADT(VAR snode: TException);


{[]====================[]}
{| Exception code names |}
{[]====================[]}

{ According to the Turbo Pascal Manual, the runtime
  errors are divided into five categories:
    - [1..99]    : DOS Errors
    - [100..149] : I/O Errors
    - [150..199] : Critical Errors
    - [200..255] : Fatal Errors
    - [256..+oo] : User Defined Errors
}

CONST
  No_Exception                  =   0; { ZERO }

  { DOS Errors: [1..99] }
  Invalid_Function_Number       =   1;
  File_Not_Found                =   2;
  Path_Not_Found                =   3;
  Too_Many_Open_Files           =   4;
  File_Access_Denied            =   5;
  Invalid_File_Handle           =   6; {...}
  Invalid_File_Access_Code      =  12; {...}
  Invalid_Drive_Number          =  15;
  Cannot_Remove_Current_Dir     =  16;
  Cannot_Rename_Across_Drives   =  17;
  No_more_files                 =  18; {...}

  { I/O Errors: [100..149] }
  Disk_Read_Error               = 100;
  Disk_Write_Error              = 101;
  Disk_Full                     = Disk_Write_Error;
  File_Not_Assigned             = 102;
  File_Not_Open                 = 103;
  File_Not_Open_for_Input       = 104;
  File_Not_Open_for_Output      = 105;
  Invalid_Numeric_Format        = 106; {...}


  { Critical Errors: [150..199] }
  Disk_is_Write_Protected       = 150;
  Unknown_Unit                  = 151;
  Bad_Drive_Request_Struct_Len  = 151;
  Drive_Not_Ready               = 152;
  Unknown_Commmand              = 153;
  CRC_Error_in_Data             = 154;
  Bad_Drive_Req_Struct_Length   = 155;
  Disk_Seek_Error               = 156;
  Unknown_Media_Type            = 157;
  Sector_Not_Found              = 158;
  Printer_Out_Of_Paper          = 159;
  Device_Write_Fault            = 160;
  Device_Read_Fault             = 161;
  Hardware_Failure              = 162; {...}

  { Fatal Errors: [200..255] }
  Division_by_Zero              = 200;
  Range_Check_Error             = 201;
  Program_Stack_Overflow        = 202;
  Stack_Overflow_Error      = Program_Stack_Overflow;
  Program_Heap_Overflow         = 203;
  Heap_Overflow_Error       = Program_Heap_Overflow;
  Invalid_Pointer_Operation     = 204;
  Floating_Point_Overflow       = 205;
  Floating_Point_Underflow      = 206;
  Invalid_Floating_Point_Opr    = 207;

  Overlay_Manager_Not_Installed = 208;
  Overlay_File_Read_Error       = 209;
  Object_Not_Initialized        = 210;
  Call_to_Abstract_Method       = 211;
  Stream_Registration_Error     = 212;
  Collection_Index_Out_of_Range = 213;
  Collection_Overflow_Error     = 214;
  Arithmetic_Overflow_Error     = 215;
  General_Protection_Fault      = 216;

  { Programmer Defined Errors: [225..255] }
  Too_Many_Nested_Exceptions    = 225;
  Exception_Stack_Overflow      = 226;
  Missing_Exception_Handler     = 227;
  Re_Throw_Without_Throw        = 228;
  UnReleased_Exception_Context  = 229;
  Object_Not_Destroyed          = 230;

  Ctrl_Break                    = 255; { Turbo's }

  { errors are numbered 0..255 because
    System.RunError() takes a BYTE argument,
    not an INTEGER. }
CONST
  Exception_Stack_Size = 32;
  { [Lee-83] says that 16 is enough, but recursive }
  {          procedures break this claim.          }

{]================[}
{} IMPLEMENTATION {}
{]================[}

{$IFDEF Ver40}
   {$DEFINE Version_Supported}
{$ENDIF}

{$IFDEF Ver50}
   {$DEFINE Version_Supported}
{$ENDIF}

{$IFDEF Ver55}
   {$DEFINE Version_Supported}
{$ENDIF}

{$IFDEF Ver60}
   {$DEFINE Version_Supported}
{$ENDIF}

{$IFDEF Ver70}
   {$DEFINE Version_Supported}
{$ENDIF}

{$IFNDEF Version_Supported}
  ERROR: Compiler version not supported!
{$ENDIF}


TYPE
  PNode = ^TNode;
  TNode = RECORD
    done : PDestructor;
    next : PNode;
    prev : PNode;
    obj  : POINTER;
  END;
{ TNode: Node within each ADT instance used to build
         the linked list of objects that will be
         destroyed when an exception occurs.

  done: Pointer the obj^'s destructor procedure.

  next && prev: Pointers used to chain all the
    exception fields in ADT instances that are
    registered in the destructor stack.

  obj: Pointer to the ADT instance that is registered
    in the destructor stack. This field could be
    spared, but for this the programmer would be
    required to put the exception field as the very
    first field in each ADT instance, which seems like
    a stiff requirement.

  This list is double linked because some programmers
  don't destroy ADTs in the reverse order of
  construction, which means that sometimes one of
  these links will be removed from the middle of the
  list. Nevertheless, when Except destroys intances it
  will do so in the reverse order of registration.
}

  { Here is how pointers are layed out in memory }
  TFar_Pointer = RECORD
    CASE INTEGER OF           { return address       }
      1:(ofs, seg: WORD;   ); { view it as 2 words   }
      2:(jump    : POINTER;); { view it as a pointer }
  END;                        { jump == seg*16 + ofs }

  { Activation record memory layout }
  TActivation_Record = RECORD
    sp : WORD;  { caller's sp register }
    bp : WORD;  { caller's bp register }
  END;

  PExcept_Record = ^TExcept_Record;
  TExcept_Record = RECORD
    sp : WORD;                { caller's sp register }
    bp : WORD;                { caller's bp register }
    ADT_Automatic : PNode;
    CASE INTEGER OF           { return address       }
      1:(ofs, seg: WORD;   ); { view it as 2 words   }
      2:(jump    : POINTER;); { view it as a pointer }
  END;

{ Global stack to store all exception records }
CONST
  top : INTEGER = -1;  { stack top }
  Stack_Bound = Exception_Stack_Size-1;
VAR
  stack : ARRAY[0..Stack_Bound] OF TExcept_Record;

{ TExcept_Record
  When an exception context is entered a new exception
  record is assigned to it from the exception stack.
  The exception record contains the required data
  to jump to the exception handler, and also records
  the list of automatic and heap varibles that should
  be destroyed before leaving the exception context.

  sp: Pointer to the begining of the stack frame for
    the procedure.

  bp: Chain of pointers to the callers stack. This is
    the pointer chain that is traversed when unwinding
    the stack.

    These two pointers are stack based, which means that
    to to get the real pointer they must be completed
    with the stack segment, as follows:
     Real_Pointer := System.Ptr(System.SSeg, sp);

  ADT_Automatic: Pointer to the first ADT instance
    allocated in the program stack that got registered
    invoking procedure Register_ADT() within the
    current exception context.
    - An ADT stack instance, also called an automatic
      variable, is easily identifiable because its
      segment address always is System.SSeg.
    - This data structure chains together all the
      instances that will be destroyed when an
      exception occurs and before the exception
      context is left.
    - Upon program abort, all registered destructors
      that correspond to currently active activation
      records are invoked.
    - This double linked list functions as a stack
      most of the time, unless the programmer destroys
      objects and then reinitializes them.
}

CONST
  GLOBAL : RECORD
    ExitSave                : POINTER;
    ErrorAddr               : POINTER;
    exception_code          : INTEGER;
    last_exception_code     : INTEGER;
    error_handler_installed : BOOLEAN
  END = (
    ExitSave                : NIL;
    ErrorAddr               : NIL;
    exception_code          : No_Exception;
    last_exception_code     : No_Exception;
    error_handler_installed : FALSE
  );  { GLOBAL }

{ ExitSave: Contains a pointer to the second exit
    procedure that should be invoked upon program
    termination (the first procedure in the chain
    always is Run_Time_Error_Handler).

  ErrorAddr: Pointer to the procedure or statement
    that caused the program to abort. This variable is
    set to a value different than NIL when a Throw(),
    Signal() or Re_Throw() is issued but there in no
    active exception context.

  exception_code: Contains the value that will be
    returned by Catch() when an exception occurs.

  last_exception_code: This variable holds the
    previous value stored in "exception_code". It is
    needed because Catch() clears "exception_code",
    but after this the programmer can request this
    same value through procedure Exception_Code().

  error_handler_installed: This boolean is initialized
    to FALSE, because the first time that function
    Exception() is executed it installs
    Run_Time_Error_Handler() as the first exit
    procedure. In this way the programmer does not
    have to put unit Except as the last unit used in a
    program, because it install itself as the first in
    the exit procedure chain.
}
PROCEDURE Register_ADT(  { EXPORT }     { ADH }
  {+} VAR obj;               { ADT instance }
  {+} VAR snode: TException; { exception field }
  {+}     done : POINTER     { ^ADT destructor }
);
{ RESULTS
  Registers "obj" in the current exception context,
  adding it to the ADT destructor stack.
  - Only registered ADT instances will get their
    destructor invoked when an exception is thrown.
  - Only objects allocated in the program stack
    (automatic objects) get registered by this
    operation. Non automatic objects, which are
    allocated in dynamic memory, are never registered,
    because the programmer is responsible for
    destroying them.
  - "snode" must be the exception field withing the
    ADT instance "obj".
  - "done" is (a pointer to) the destructor for "obj".

  - The following code should be used to register and
    ADT intance:
      Except.Register_ADT(ADT, ADT.Rep.excp, @Done);
    where "excp" is the exception field within the
    ADT instance "ADT" and "Done" is (a pointer to)
    the destructor operation.
  - This operation should be invoked at the end of the
    constructor, to prevent the destruction of an
    instance that was not completely constructed.

  - When no exception context is active nothing
    happens.
  - When an exception occurs, the destructors for all
    registered automatic ADT instances will be invoked
    before leaving the invocation of each procedure.
  - It is incorrect to destroy and object to later
    reuse it, because it might get chained to the
    incorrect exception context. This means that the
    Clear() operation in an ADT cannot be programmed
    as Done() followed by Init(). }
BEGIN { Register_ADT }
  { Only register an automatic instances }
  IF System.SSeg <> System.Seg(obj) THEN BEGIN
    EXIT;
  END;
  { Only register in an active exception context }
  IF top < 0 THEN BEGIN
    EXIT;
  END;

  TNode(snode).done := PDestructor(done);
  TNode(snode).obj  := ADDR(obj);

  TNode(snode).next :=
    stack[top].ADT_Automatic;
  TNode(snode).prev := NIL;

  { chain back }
  IF stack[top].ADT_Automatic <> NIL THEN BEGIN
     stack[top].ADT_Automatic^.prev :=
       Pnode( ADDR(snode) );
  END;

  Register_ADT(obj, snode, @UnRegister_ADT);

  stack[top].ADT_Automatic := Pnode( ADDR(snode) );
END;  { Register_ADT }

PROCEDURE Invoke_Destructors(       { LOCAL } { ADH }
  {?} VAR context: PNode { stack[top].ADT_Automatic }
);
{ RESULT
  Invokes the destructor for every registered ADT in
  the current exception context.
  - The excepcion stack must not be empty.
  - context should be stack[top].ADT_Automatic.     }
BEGIN { Invoke_Destructors }
  WHILE (context <> NIL) DO BEGIN
    { destroy this object }
    context^.done(context^.obj^);

    { work up the stack }
    context := context^.next;
  END;

{ If the client programmer "chained" the destructors
  for its ADT instances, it is possible that for value
  of stack[top].ADT_Automatic to change after a
  destructor invocation. "Chaining" here means that
  the destructor for one (automatic) object destroys
  another (automatic) object.

  In this case, as "context" is the destructor chain
  head and not a copy of its initial value, such an
  event will not break this code. }
END;  { Invoke_Destructors }

PROCEDURE UnRegister_ADT(   { EXPORT }  { ADH }
  {+} VAR snode: TException { exception field }
);
{ RESULTS
  Unlinks the ADT instance that contains "snode" from
  the ADT Destructor stack.
  - This operation should be called once for each ADT
    instance that got chained in the Destructor stack.
  - Usually, this operation is called in the body of
    the destructor as follows:
      Except.UnRegister_ADT(ADT.Rep.excp); }
BEGIN { UnRegister_ADT }
  { Only register an automatic instances }
  IF System.SSeg <> System.Seg(snode) THEN BEGIN
    EXIT;
  END;
  { Only unregister from an active exception context }
  IF top < 0 THEN BEGIN
    EXIT;
  END;

  IF TNode(snode).prev = NIL THEN BEGIN
    { unlink the stack top }
    stack[top].ADT_Automatic := TNode(snode).next;
  END
  ELSE BEGIN
    { unlink in the middle of the chain }
    TNode(snode).      prev^.next :=
      TNode(snode).                  next;

    IF TNode(snode).next <> NIL THEN BEGIN
       TNode(snode).   next^.prev :=
       TNode(snode).                 prev;
    END;
  END;  
END;  { UnRegister_ADT }

FUNCTION Normalized(                 { David }
  {+}   p : POINTER   { pointer to normalize }
) {>>>>>} : POINTER;
{ RESULTS
  Returns p as a normalized pointer.         }
InLine( { Normalized }
  $58/          { pop ax    ;pop offset into ax  }
  $5A/          { pop dx    ;pop segment into dx }
  $89/$C3/      { mov bx,ax ;bx = ofs(p^)        }
  $B1/$04/      { mov cl,4  ;cl = 4              }
  $D3/$EB/      { shr bx,cl ;bx = ofs(p^) div 16 }
  $01/$DA/      { add dx,bx ;add bx to segment   }
  $25/$0F/$00   { and ax,$F ;mask out unwanted   }
);      { Normalized }      {  bits in offset    }

PROCEDURE Long_Jump(                  { David }
  {+} VAR ex : TExcept_Record { where to jump }
);
{ RESULTS
  Restores the BP & SP register from "ex",
  and then does a far jump to "ex.jump".
  - It never returns.                        }
InLine( { Long_Jump }
  $5F/             { pop di            ;di = ofs(ex) }
  $07/             { pop es            ;es = seg(ex) }
  $26/$8B/$25/     { mov sp, [es:di]   ;restore sp   }
  $26/$8B/$6D/$02/ { mov bp, [es:di+2] ;restore bp   }
  $B8/$01/$00/     { mov ax,=1         ;Excp := TRUE }
  $26/$FF/$6D/$08  { jmp far [es:di+8] ;go to ex.jump}
);      { Long_Jump }

PROCEDURE Set_Exception_Code(             { ADH }
  {+} code : INTEGER       { new exception code }
);
{ RESULTS
  Changes GLOBAL.exception_code, and remembers the 
  previous code for later use by Re_Throw.  }
BEGIN { Set_Exception_Code }
  GLOBAL.last_exception_code := GLOBAL.exception_code;
  GLOBAL.exception_code      := code;
END;  { Set_Exception_Code }

PROCEDURE Goto_Except_Record(                 { ADH }
  {+} VAR ex   : TExcept_Record; { exception record }
  {+}     code : INTEGER         { exception code   }
);
{ RESULTS
  Sets the exception code to "code", and does a
  Long_Jump to ex.jump, restoring the activation
  record for that invoking procedure.
  - This means that the current program stack is 
    popped all the way to the point pointed to by 
    "ex". Many procedure invocations will not return
    where they should, because the next instruccion to 
    be executed will be that of the exception context 
    pointed to by "ex".  }
BEGIN { Goto_Except_Record }
  Set_Exception_Code(code);
  Long_Jump(ex);
END;  { Goto_Except_Record }

FUNCTION Exception_Code  { EXPORT }     { ADH }
  {>>>>} : INTEGER;
{ RESULT
  Returns the last exception code.
  - This function can be called many times, as it does
    not clear the exception code. On the contrary,
    Catch() should be called only once, because it
    does clear the exception code. }
BEGIN { Exception_Code }
  Exception_Code := GLOBAL.last_exception_code;
END;  { Exception_Code }

FUNCTION Catch           { EXPORT }     { ADH }
  {>>>} : INTEGER;
{ RESULT
  Returns the last exception code, and clears it. }
{ REQUIRES
  This function should be called only once, as it
  clears the exeception code.
  - To retrieve again the last exception code,
    function Exception_Code can be used.  }
BEGIN { Catch }
  GLOBAL.last_exception_code := GLOBAL.exception_code;
  Catch                      := GLOBAL.exception_code;
  GLOBAL.exception_code      := No_Exception;
END;  { Catch }

{$F+  Use Far Call memory model }

FUNCTION Heap_Error_Function( { BORLAND }
  {+} size : WORD        { ignored here }
) {>>>>>>} : INTEGER;
{ RESULTS
  Forces heap allocation errors to abort the program }
{ REMARKS
  When a heap error occurs, because the heap manager 
  cannot complete an allocation request, the Pascal
  Run Time system calls the procedure pointed to by 
  variable System.HeapError.

  The default heap error function returns a cero, 
  which results in a program abort. The programmer can
  replace it with a smarted one, executing the
  following code:
  - System.HeapError := @Heap_Error_Function;

  This procedure is never called in this unit, but it
  is included here to remind the hacker how to further
  control this type of exceptions. The manual provides
  more information on this respect.  }
BEGIN { Heap_Error_Function }
  { 0 is the default: forces a runtime error }
  Heap_Error_Function := 0;
END;  { Heap_Error_Function }

{$F- Restore default }

PROCEDURE Register_Exit_Procedure( { EXPORT }  { ADH }
  {?} VAR ExitSave : PExitProc;  { Exit Chain Ptr }
  {+}     ExitProc : PExitProc   { Exit procedure }
);
{ RESULTS
  Installs ExitProc as a Turbo Pascal Exit procedure.
  For this, ExitProc gets inserted into the Exit
  chain.
  - Upon return, ExitSave is the POINTER to the exit
    procedure that should be called after the
    installed exit procedure returns.
  - Upon its execution, the installed Exit procedure
    must chain to other exit procedures, following the
    protocol described in the Turbo Pascal manual.
    This is accomplished restoring System.ExitProc to
    the value returned by this procedure in ExitSave.
  - Once installed, and exit procedure can never be
    removed from the exit chain.  }
BEGIN { Register_Exit_Procedure }
  { Chain ExitProc to be the procedure that gets }
  { executed after Run_Time_Error_Handler        }

  ExitSave := PExitProc( GLOBAL.ExitSave );

  PExitProc( GLOBAL.ExitSave ) := ExitProc;

END;  { Register_Exit_Procedure }

{$F+ Use far call model }

PROCEDURE Run_Time_Error_Handler;              { ADH }
{ RESULTS
  Catches all runtime errors, and invokes an error 
  handler if one exists. Otherwise, it simple lets the 
  program abort.
  - Regretfully, Program_Stack_Overflow's cannot 
    always be detected, because sometimes there is no 
    stack where to work. But more often than not the
    Program_Stack_Overflow exception is detected and 
    well handled.  }
{ NOTES
  As this routine does use local variables, when a
  Program_Stack_Overflow occurs there is a chance that
  the program stack would get corrupted, with fatal
  consecuences throughout.
  - Nothing is perfect. }
BEGIN { Run_Time_Error_Handler }
  IF  (top >= 0)  THEN BEGIN
  { On exit Turbo sets System.ExitProc to NIL before
    calling every exit procedure. As this procedure is
    called only when Turbo tries to finish up the
    program, to catch the next exception the run time
    error handler must be reinstalled.  }

  { RunError always sets System.ErrorAddr to NIL }

    IF System.ExitProc = NIL THEN BEGIN
      { Set again System.ExitProc to }
      { catch the next exception     }
      System.ExitProc := @Run_Time_Error_Handler;
    END;

    { invoke the destructor for every ADT instance }
    { registered in the curren exception context   }
    Invoke_Destructors(stack[top].ADT_Automatic);

    DEC(top);

    GLOBAL.last_exception_code:=GLOBAL.exception_code;
    GLOBAL.exception_code     :=System.ExitCode;

    Goto_Except_Record(stack[top+1], System.ExitCode);
    { the exception has been caught!!!    }
  END;

{ If it gets here it means that there are no more
  exception contexts. There is no choice but to let
  the program abort.  }

  IF (GLOBAL.exception_code=Missing_Exception_Handler)
  THEN BEGIN
  { The exception stack was empty, and the Missing
    Exception Handler exception was somehow issued,
    probably by Throw(), Signal() or Re_Throw().
    - To blame whoever threw the exception,
      System.ErrorAddr is restored.               }
    System.ErrorAddr := GLOBAL.ErrorAddr;
    System.ExitCode  := Missing_Exception_Handler;
  END;

  { Now before termination, we chain }
  { to other unit's exit procedures  }
  System.ExitProc :=  GLOBAL.ExitSave;
END;  { Run_Time_Error_Handler }

{$F- restore default }

FUNCTION Exception: BOOLEAN; { EXPORT } { ADH }
{ RESULTS
  Sets an exception mark in the  exception stack.
  - Upon execution of the Throw() or Signal()
    procedures, this mark will be used to restore the
    program's state to a previous activation of
    Exception.
  - Unlike a normal routine call, Throw() will branch
    to where Exception would have returned when called
    the first time.
  - Exception always returns TRUE, but when its 
    invocation comes from Throw() it will return
    FALSE.  In this way, the calling procedure can 
    tell whether an exception has ocurred, and handle 
    it accordingly.
  - Exception sets GLOBAL.exception_code to
    No_Exception, and saves the current SP, BP, and a 
    jump destination.
  - Exception cannot be invoked from the 
    initialization code of any unit, unless unit 
    Except is the last unit used in the main program.
  - The programmer is always responsible for 
    destroying in every exception handler every 
    instance that was contructed within the exception 
    context.

> COST
  The cost of calling Exception is always O(1).
  - Any penalties for exception handling are paid when 
    an exception occurs. However, the worst is O(n), 
    where "n" is the depth of the stack.
}
VAR
  go_on : BOOLEAN;
  caller: TExcept_Record; { caller's activation rec }
BEGIN { Exception }
{
First use InLine to get Exception's return address. 
Later, when an exception is thrown, a return will be
made to this return address, but the value returned 
will be TRUE.

Exception returns FALSE, but when Long_Jump'in into 
it, it returns TRUE.


      Runtime program's stack memory layout
      =====================================

 THE PROGRAM STACK GROWS FROM HIGH MEM TO LOW MEM!

00000   (Low mem)          This is a picture of
       +-------------+     the program stack at
SP  -> |+++++++++++++|     runtime. It is in here
       |  Exception  |     where all local
       |+++++++++++++|     variables are stored.
       +-------------+                           ^
BP --> |     BP      |--+  When a procedure is  /|\
       +-------------+  |  called, the stack     |
       |   ret ofs   |  |  grows toward low mem. |
       |   ret seg   |  |                        |
       +-------------+  |  When the procedure
CSP -> |/////////////|  |  returns, the stack    |
       |caller's VARs|  |  is poped out toward   |
CBP -> |/////////////|<-+  high mem.            \|/
       +-------------+                           v
                             CBP = Caller's BP
FFFFF   (High mem)           CSP = Caller's SP

This picture is similar to those from the Borland C++ 
programmer's manuals. C++ supports more memory models 
than Pascal which means that its manuals are far more 
informative than the Turbo Pascal manuals.

The slashed area is the caller's stack area, where the 
caller's local variables are. The area marked with +'s 
is the section of the stack where the local variables 
for function Exception are stored.

Exception must extract from the current program stack 
the Exception record for its caller, to store it in 
the Exception stack. The exception record for any 
procedure invocation consists of two parts:
  1- The Activation record
  2- The procedure return address
In the case of the Intel x86 architecture, the 
Activation record for any procedure invocation consist 
of it's BP and SP register values.

Function Exception must extract these two components 
from the program stack. As Turbo doesn't have
procedures to walk within the program stack, in this
implementation it is necessary to use Assembly 
Language to peek into it.

As it can be seen from the above diagram, the BP for a 
calling procedure is always bigger than the BP for the
called procedure. In other words, the BP for the
called procedure is always less than the calling's 
procedure BP. This means that the BPs stored in the 
exception stack are in decreasing order.

}

InLine(
{ The following code extracts the Exception_Record for
  whoever invoked Exception. Assembly language is used
  for this because the activation record for any
  procedure consist of it's BP and SP pointers, which
  are stored in the program stack. }

  { == Entry code, generated by Turbo Pascal == }
  {/$55             push bp                     }
  {/$8B/$EC         mov  bp, sp                 }
  {/$2B/$06/Local   sub  ax, SizeOf(Local_Vars) }

  { Stores in "caller" the caller's sp, bp and  }
  { Exception's current return address.         }
   $8B/$46/$00      { mov ax, [bp]     ;             }
  /$89/$86/caller+2 { mov caller.bp,ax ;cllr.bp:=CBP }

  /$8B/$C5          { mov ax, bp       ;BP-6 == CSP  }
  /$05/$06/$00      { add ax, =6       ;             }
  /$89/$86/caller+0 { mov caller.sp,ax ;cllr.sp:=CSP }
  { =6 == SizeOf(BP)+SizeOf(ret.ofs)+SizeOf(ret.seg) }

  { Copies the return address to new bp && to caller }
  /$8b/$46/$02      { mov ax, [bp]+2    ;ret.ofs     }
  /$89/$86/caller+8 { mov caller.ofs,ax ;save it     }
  /$8B/$46/$04      { mov ax, [bp]+4    ;ret.seg     }
  /$89/$86/caller+10{ mov caller.seg,ax ;save it     }

{ As it can be seen from the above code, Exception is
  quite an inexpensive operation, which makes
  exception handling cheap.
  - In this implementation the price of handling
    exceptions is payed later, when an exception is
    thrown, and not when an exception context is
    entered.  }
);

  { Check for at least one invalid exception context }
  IF (top >= 0) THEN
    IF (stack[top].bp < caller.bp) THEN BEGIN
  { The programmer forgot to put at least one
    invocation to Leave_Exception_Context for an
    exception context. This means that some of the
    Exception records in the exception stack are no
    longer valid: the code in here tries to remove as
    many of those as possible. Nonetheles, the program
    is incorrect, and should be fixed.  }
    go_on := TRUE;
    WHILE go_on DO BEGIN
    { The topmost entries belong to procedures that
      have already returned, so their exception
      records are removed from the stack.  }

      DEC(top);
      IF top >= 0 THEN BEGIN
        go_on := (stack[top].bp < caller.bp);
      END
      ELSE BEGIN
        go_on := FALSE;
      END;
    END;

    { Now abort the current procedure, by faking a  }
    { Throw(): this is easier than Long_Jump()'ing. }
    Set_Exception_Code(UnReleased_Exception_Context);
    Exception := TRUE;
    EXIT;
  END;

  IF top = Stack_Bound THEN BEGIN
  { The exception stack is full:
    break out and blame the caller }
    Set_Exception_Code(Exception_Stack_Overflow);

    { fake a Throw: it's easier than Long_Jump'ing }
    Exception := TRUE;
    EXIT;
  END;

{ Now comes the trick to avoid forcing the programmer
  to use Except as the last UNIT in the program.  }

{ Make sure that upon program termination Except's
  Run_Time_Error_Handler() is called before any other
  procedure, even when Except is not the last UNIT
  used in a program. This code also lets the
  programmer to chain newer Exit procedures into the
  Exit chain. However, no procedure should ever be
  removed form the exit chain, because what would get
  removed would be the Run_Time_Error_Handler!  }

  IF NOT GLOBAL.error_handler_installed THEN BEGIN
    GLOBAL.ExitSave := System.ExitProc;
    System.ExitProc := @Run_Time_Error_Handler;

    GLOBAL.error_handler_installed := TRUE;
  END;

  { Push the exception record in the exception stack }
  INC(top);
  stack[top] := caller;
  stack[top].ADT_Automatic := NIL;

  { Make sure that nobody else reports run time errors }
  System.ErrorAddr := NIL;

  Set_Exception_Code(No_Exception);
  Exception := FALSE;

  { ==== Standard Turbo Pascal exit code ====  }
  { InLine(                                    }
  {  $8B/$E5        mov sp, bp ; for this proc }
  { /$5D            pop bp     ;  there are no }
  { /$CA/$00/$00);  ret far 0  ;  parameters   }
END;  { Exception }

{
 000 low mem    ^                |
    +--------+ /|\               |
    | Proc_C |  |  S    Proc_A;  |  t
    +--------+  |  t             |  i
    | Proc_B |  |  a    Proc_B;  |  m
    +--------+  |  c             |  e
    | Proc_A |  |  k    Proc_C;  |
    +--------+  |               \|/
 FFF High mem   |                v


  SP(Proc_C) < SP(Proc_B) < SP(Proc_A)

  SP(Proc_A) > SP(Proc_B) > SP(Proc_C)


  Proc_A called Proc_B ==> SP(Proc_A) > SP(Proc_B)

  Proc_B called Proc_C ==> SP(Proc_B) > SP(Proc_C)


  Proc_C returns to Proc_B ==> SP(Proc_C) < SP(Proc_B)

  Proc_B returns to Proc_A ==> SP(Proc_B) < SP(Proc_A)
}

PROCEDURE Throw(         { EXPORT }     { ADH }
  {+} code : INTEGER     { exception code }
);
{ RESULTS
  Generates an exception by branching to the enclosing
  exception handler and returning the "code" as the
  exception code.
  - The enclosing context need not be in an invoking
    procedure: Throw() can branch to an exception
    context that where the current exception context
    is nested.
  - All destructors for automatic objects registered
    in the current exception context are invoked
    before branching to the exception handler.
  - The programmer is always responsible for
    destroying all dynamic objects, which are those
    allocated in dynamic memory.
  - Restores SP, BP, and then jumps. }
VAR
  procs_bp : WORD;
  ex       : TExcept_Record;
  go_on    : BOOLEAN;
  seg, ofs : WORD;
BEGIN { Throw }
  { get the procedure's bp }
  InLine(
    $8B/$46/$00/     { mov ax, [bp]     }
    $89/$86/procs_bp { mov procs_bp, ax }
  );
{ procs_bp is the the BP for the procedure that
  is Raising the exception.                     }

{ Pop from the exception stack every exception record
  that corresponds to the invocation of a procedure
  that already returned. These are the records whose
  bp is deeper in the stack, this is, those that have
  a smaller bp than the current bp.  }

  go_on := (top >= 0);
  WHILE go_on DO BEGIN
    IF (stack[top].bp < procs_bp) THEN BEGIN
    { This code should never be executed. It
      corresponds to exception contexts that the
      programmer didn't release properly invoking
      Leave_Exception_Context.  }
      DEC(top);
      go_on := (top >= 0);
    END
    ELSE BEGIN
      go_on := FALSE;
    END;
  END;
  { Note that stack[].bp should be LESS than the }
  { caller's bp. Signal tests for  LESS EQUAL <= }


  { check for a missing exception context }
  IF top < 0 THEN BEGIN
    { Saves the caller's return address }
    Inline(
       $8b/$46/$02  { mov ax,  [bp]+2 ;ofs }
      /$89/$86/ofs  { mov ofs, ax          }

      /$8B/$46/$04  { mov ax,  [bp]+4 ;seg }
      /$89/$86/seg  { mov seg, ax          }
    );

    { Blame Throw() for the runtime error }
    GLOBAL.ErrorAddr := System.Ptr(seg,ofs);

  { RunError sets System.ErrorAddr to NIL and
    System.ExitCode to "status". The use of a GLOBAL
    is necesary to inform procedure
    Run_Time_Error_Handler that the exception stack is
    empty.  }

    System.RunError(Missing_Exception_Handler);
  END;

  Invoke_Destructors(stack[top].ADT_Automatic);
  ex := stack[top];
  DEC(top);
  Goto_Except_Record(ex, code);
END;  { Throw }

PROCEDURE Signal(        { EXPORT }     { ADH }
  {+} code : INTEGER     { exception code }
);
{ RESULT
  Returns control to an exception handler, making sure
  it is not in the same procedure where Signal is
  invoked. If many exception blocks have been nested
  in the same procedure, then all of them would be
  poped off the exceptions stack.
  - The destructors for those ADTs intances registere
    in each exception context are invoked before
    leaving the exception context.  }
{ RESULTS
  Generates an exception by branching to the enclosing
  exception handler and returning the "code" as the 
  exception code.
  - The enclosing context need not be in an invoking 
    procedure: Throw() can branch to an exception
    context that where the current exception context
    is nested.
  - All destructors for automatic objects registered
    in the current exception context are invoked
    before branching to the exception handler.
  - The programmer is always responsible for
    destroying all dynamic objects, which are those
    allocated in dynamic memory.
  - Restores SP, BP, and then jumps.        }
VAR
  procs_bp : WORD;
  ex       : TExcept_Record;
  go_on    : BOOLEAN;
  seg, ofs : WORD;
BEGIN { Signal }
  { get the caller's bp }
  InLine(
    $8B/$46/$00/     { mov ax, [bp]     }
    $89/$86/procs_bp { mov procs_bp, ax }
  );

  { Pop off every valid exception record that has }
  { the same bp as the caller, ie. those that     }
  { belong in the same procedure. This loop       }
  { allows for nested exception blocks inside a   }
  { single procedure, which should be rare.       }

  go_on := (top >= 0);
  WHILE go_on DO BEGIN
    IF (stack[top].bp < procs_bp) THEN BEGIN
      { These are exception records that the
        programmer forgot to clear. The program
        is incorrect, and should be fixed. }
      DEC(top);
      go_on := (top >= 0);
    END
    ELSE IF (stack[top].bp = procs_bp) THEN BEGIN
      { These top most entries belong to the
        same procedure: they are ignored.  }
      Invoke_Destructors(stack[top].ADT_Automatic);
      DEC(top);
      go_on := (top >= 0);
    END
    ELSE BEGIN
      { stack[top].bp < procs_bp ==> there exists
        an exception context that does not belong
        in the current procedure. }
      go_on := FALSE;
    END;
  END;

  { check for a missing exception context }
  IF top < 0 THEN BEGIN
    { Saves the caller's return address }
    Inline(
       $8b/$46/$02  { mov ax,  [bp]+2 ;ofs }
      /$89/$86/ofs  { mov ofs, ax          }

      /$8B/$46/$04  { mov ax,  [bp]+4 ;seg }
      /$89/$86/seg  { mov seg, ax          }
    );

    { Blame Signal() for the runtime error }
    GLOBAL.ErrorAddr := System.Ptr(seg,ofs);
    System.RunError(Missing_Exception_Handler);
  END;

  Invoke_Destructors(stack[top].ADT_Automatic);
  ex := stack[top];
  DEC(top);
  Goto_Except_Record(ex, code);
END;  { Signal }

PROCEDURE Re_Throw;      { EXPORT }     { ADH }
{ RESULTS
  Throws again the exception, usign the previous
  exception code.
  - Re_Throw() should only be used within an exception
    handler.  In some cases Re_Throw() can tell when
    this is not the case, but in general it will jump
    away to the closest exception handling context,
    with the wrong exception status code.
  - The programmer is responsible of destroying every
    ADT instances that was constructed whithin the
    current exception context. }
VAR
  ex  : TExcept_Record;
  seg,
  ofs : WORD;
BEGIN { Re_Throw }
  { check for a missing exception context }
  IF top < 0 THEN BEGIN
    { Saves the caller's return address }
    Inline(
       $8b/$46/$02  { mov ax,  [bp]+2 ;ofs }
      /$89/$86/ofs  { mov ofs, ax          }

      /$8B/$46/$04  { mov ax,  [bp]+4 ;seg }
      /$89/$86/seg  { mov seg, ax          }
    );

    { Blame Re_Throw() for the runtime error }
    GLOBAL.ErrorAddr := System.Ptr(seg,ofs);
    System.RunError(Missing_Exception_Handler);
  END;

  Invoke_Destructors(stack[top].ADT_Automatic);
  ex := stack[top];
  DEC(top);
  Goto_Except_Record(ex, GLOBAL.last_exception_code);
END;  { Re_Throw }

PROCEDURE Leave_Exception_Context;  { EXPORT } { ADH }
{ RESULTS
  Releases the last valid exception handler. }
{ REQUIRES
  Leave_Exception_Context should be the very last 
  statement in a code block that has an exception
  handler. It should also be used before any call to 
  the EXIT procedure. (EXIT is Turbo Pascal's RETURN 
  statement).

  Every call to Exception must contain a matching call 
  to Leave_Exception_Context.
  - Failing to do this will put out of synch the 
    exception stack with the program stack, and the 
    result could be that and invalid exception context 
    gets activated. This is far more than disastrous, 
    because the program will almost for sure hang the 
    machine.  }

{ NOTES
  This procedure is necesary because the exception
  package is an add-on to Pascal. This procedure would 
  be unnecesary if the compiler would generate code to
  remove the last valid handler whenever an exception
  block was left. Regretfully, there is no way to tell
  when an exception block has been exited. This
  restriction is what makes makes exception handling
  trickier to use, at least to the novice.  }
VAR
  seg,
  ofs : WORD;
BEGIN { Leave_Exception_Context }
  IF top < 0 THEN BEGIN
    { There's no active current exception context }
    Inline(
       $8b/$46/$02  { mov ax,  [bp]+2 ;ofs }
      /$89/$86/ofs  { mov ofs, ax          }

      /$8B/$46/$04  { mov ax,  [bp]+4 ;seg }
      /$89/$86/seg  { mov seg, ax          }
    );

    { Blame Leave_Exception_Context() for the error }
    GLOBAL.ErrorAddr := System.Ptr(seg,ofs);
    System.RunError(Missing_Exception_Handler);
  END;

  IF (stack[top].ADT_Automatic <> NIL) THEN BEGIN
    { the programmer forgot to destroy all instances }
    Inline(
       $8b/$46/$02  { mov ax,  [bp]+2 ;ofs }
      /$89/$86/ofs  { mov ofs, ax          }

      /$8B/$46/$04  { mov ax,  [bp]+4 ;seg }
      /$89/$86/seg  { mov seg, ax          }
    );
    GLOBAL.ErrorAddr := System.Ptr(seg,ofs);
    System.RunError(Object_Not_Destroyed);
  END;

  { Pop off the last exception record in the stack }
  DEC(top);
END;  { Leave_Exception_Context }

FUNCTION Exception_Depth { EXPORT }     { ADH }
{>>>>>>>}: INTEGER;
{ RESULTS
  Returns the number of active exception contexts. }
BEGIN { Exception_Depth }
  Exception_Depth := top+1;
END;  { Exception_Depth }

BEGIN { Except }
END.  { Except }

{ EOF: Except.pas }
